package magic_squares.sudoku

import magic_squares.{Displayable, Grid}

import scala.collection.mutable.ArrayBuffer

abstract class SudokuSolver(initialGrid: Grid) extends Displayable {
  protected var _initial: Grid = initialGrid
  protected var grid: Grid = copy(_initial)
  protected val SIZE: Int = SudokuSolver.SIZE
  protected val SQUARE_SIZE: Int = SudokuSolver.SQUARE_SIZE
  protected val DEBUG: Boolean = false
  protected val NUMBERS: Array[Int] = Array.range(1, SIZE + 1)

  protected def print(grid: Grid): Unit = {
    for (y: Int <- 0 until SIZE) {
      println(grid(y).mkString(","))
    }
  }

  protected def copy(grid: Grid): Grid = {
    return grid.map(_.clone())
  }

  def solve(): Unit = {
    val solved: Boolean = _solve()
    if (solved) {
      display(grid)
    } else {
      println("No solution")
    }
  }

  private def _solve(): Boolean = {
    display(grid)

    if (DEBUG) print(grid)

    val (x: Int, y: Int) = getNextEmpty()
    if (x == -1) {
      if (DEBUG) println("  Found solution")
      return true
    }
    val numbers: Array[Int] = getPossibleNumbers(x, y)
    if (numbers.isEmpty) {
      return false
    }

    if (DEBUG) println(s"  Values to test: " + numbers.mkString("[", ", ", "]"))

    for (n: Int <- numbers) {
      if (DEBUG) println(s"    Testing $n")
      grid(y)(x) = n
      if (_solve()) {
        if (DEBUG) println("    Found solution, collapsing call stack")
        return true
      }
    }
    grid(y)(x) = 0

    if (DEBUG) println(s"  No solution for this configuration")
    return false
  }

  private def getNextEmpty(): (Int, Int) = {
    for (y: Int <- 0 until SIZE) {
      for (x: Int <- 0 until SIZE) {
        if (grid(y)(x) == 0) {
          return (x, y)
        }
      }
    }
    return (-1, -1)
  }

  private def getPossibleNumbers(x: Int, y: Int): Array[Int] = {
    val inGrid: ArrayBuffer[Int] = grid(y).concat(grid.map(_(x))).to(ArrayBuffer)
    val sx: Int = (x / 3) * 3
    val sy: Int = (y / 3) * 3
    for (dy: Int <- 0 until SQUARE_SIZE) {
      for (dx: Int <- 0 until SQUARE_SIZE) {
        inGrid.addOne(grid(sy + dy)(sx + dx))
      }
    }
    return NUMBERS.diff(inGrid.distinct)
  }
}

object SudokuSolver {
  val SIZE: Int = 9
  val SQUARE_SIZE: Int = SIZE / 3

  def main(args: Array[String]): Unit = {
    val solver1: SudokuSolver = new SudokuSolver(Array(
      Array(5,3,0, 0,7,0, 0,0,0),
      Array(6,0,0, 1,9,5, 0,0,0),
      Array(0,9,8, 0,0,0, 0,6,0),

      Array(8,0,0, 0,6,0, 0,0,3),
      Array(4,0,0, 8,0,3, 0,0,1),
      Array(7,0,0, 0,2,0, 0,0,6),

      Array(0,6,0, 0,0,0, 2,8,0),
      Array(0,0,0, 4,1,9, 0,0,5),
      Array(0,0,0, 0,8,0, 0,7,9),
    )) with GraphicsDisplay
    solver1.solve()
  }
}